package auth

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"net/http"
	"strings"
	"time"

	"github.com/Masterminds/squirrel"
	openapi_errors "github.com/go-openapi/errors"
	"github.com/jmoiron/sqlx"
	"github.com/patrickmn/go-cache"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"golang.org/x/crypto/bcrypt"
	"gopkg.in/dgrijalva/jwt-go.v3"
	"orus.io/orus-io/rednerd/database"
	"orus.io/orus-io/rednerd/models"
)

var (
	// ErrInvalidCredentials is returned when auth fails because the credentials are
	// invalid
	ErrInvalidCredentials = errors.New("invalid credentials")

	// ErrAccountidCobrandidIncompatible is returned when auth fails because the
	// credentials contains both a accountid and cobrandid. This error will
	// disappear in a later version
	ErrAccountidCobrandidIncompatible = errors.New(
		"credentials cannot contain both accountid and cobrandid")
)

// TokenOptions holds options related to the JWT
type TokenOptions struct {
	Secret             string             `long:"token-secret" ini-name:"token-secret" required:"true" description:"HMAC secret for signing/verifying JWT"`
	ExpirationOpt      func(string) error `long:"token-expiration" ini-name:"token-expiration" required:"false" description:"Expiration time of the generated tokens" default:"5m"`
	Expiration         time.Duration      `no-flag:"t"`
	CookieDomain       string             `long:"cookie-domain" ini-name:"cookie-domain" description:"The domain used in cookies, must match the public-facing server host name"`
	CookieSecure       bool               `long:"cookie-secure" ini-name:"cookie-secure" description:"Set to true if the cookie must be over https only (recommended)"`
	CacheMaxSize       int                `long:"token-cache-max-size" ini-name:"token-cache-max-size" required:"faise" description:"Maximum number of entries in the token cache" default:"10000"`
	CachePurgeDelay    time.Duration      `no-flag:"t"`
	CachePurgeDelayOpt func(string) error `long:"token-cache-purge-delay" ini-name:"token-cache-purge-delay" required:"false" description:"Delay between token cache purges" default:"10m"`
}

func durationOption(tgt *time.Duration) func(string) error {
	return func(duration string) error {
		d, err := time.ParseDuration(duration)
		if err != nil {
			return err
		}
		*tgt = d
		return nil
	}
}

// NewTokenOptions creates a TokenOptions
func NewTokenOptions() *TokenOptions {
	options := TokenOptions{}
	options.ExpirationOpt = durationOption(&options.Expiration)
	options.CachePurgeDelayOpt = durationOption(&options.CachePurgeDelay)
	return &options
}

// Auth handles token authentication & creation
type Auth struct {
	db           *sqlx.DB
	secret       []byte
	cacheMaxSize int
	expiration   time.Duration
	cookieDomain string
	cookieSecure bool

	parser jwt.Parser

	log zerolog.Logger

	cache              *cache.Cache
	collectors         []prometheus.Collector
	cacheHitCollector  prometheus.Counter
	cacheMissCollector prometheus.Counter
}

// NewAuth creates a Auth instance
func NewAuth(options TokenOptions, db *sqlx.DB, log zerolog.Logger) *Auth {
	if options.CacheMaxSize == 0 {
		options.CacheMaxSize = 100
	}
	if options.Expiration == 0 {
		options.Expiration = time.Minute * 5
	}
	auth := Auth{
		db,
		[]byte(options.Secret),
		options.CacheMaxSize,
		options.Expiration,
		options.CookieDomain,
		options.CookieSecure,
		jwt.Parser{ValidMethods: []string{"HS256"}},
		log,
		cache.New(options.Expiration, options.CachePurgeDelay),
		nil,
		prometheus.NewCounter(
			prometheus.CounterOpts{
				Name: "rednerd_auth_token_cache_hit",
				Help: "The token cache hits",
			},
		),
		prometheus.NewCounter(
			prometheus.CounterOpts{
				Name: "rednerd_auth_token_cache_miss",
				Help: "The token cache misses",
			},
		),
	}

	auth.collectors = []prometheus.Collector{
		prometheus.NewGaugeFunc(
			prometheus.GaugeOpts{
				Name: "rednerd_auth_token_cache_size",
				Help: "The number of entries in the token cache",
			}, func() float64 {
				return float64(auth.cache.ItemCount())
			},
		),
		prometheus.NewGaugeFunc(
			prometheus.GaugeOpts{
				Name: "rednerd_auth_token_cache_size_max",
				Help: "The maximum number of entries in the token cache",
			}, func() float64 {
				return float64(auth.cacheMaxSize)
			},
		),
		auth.cacheHitCollector,
		auth.cacheMissCollector,
	}

	for _, c := range auth.collectors {
		prometheus.MustRegister(c)
	}

	return &auth
}

// Close release the underlying resources
func (auth *Auth) Close() {
	for _, c := range auth.collectors {
		prometheus.Unregister(c)
	}
}

func (auth *Auth) addToCache(token string, principal *models.Principal) {
	if auth.cache.ItemCount() == auth.cacheMaxSize {
		return
	}
	auth.cache.Set(token, principal, time.Until(time.Unix(principal.StandardClaims.ExpiresAt, 0)))
	if auth.cache.ItemCount() == auth.cacheMaxSize {
		// The cache just got full, sending some information if really full
		auth.cache.DeleteExpired()
		if auth.cache.ItemCount() == auth.cacheMaxSize {
			userCounters := map[string]int{}
			for _, v := range auth.cache.Items() {
				p := v.Object.(*models.Principal)
				u := userCounters[p.Subject]
				userCounters[p.Subject] = u + 1
			}
			userCountersNew := map[string]int{}
			for id, count := range userCounters {
				if count > auth.cacheMaxSize/100 {
					userCountersNew[id] = count
				}
			}
			auth.log.Warn().Dict("counters", zerolog.Dict().
				Int("total", auth.cache.ItemCount()).
				Interface("users", userCountersNew),
			).Msg("auth token cache is full")
		}
		return
	}
}

// Authenticate authenticates a user
func (auth *Auth) Authenticate(
	ctx context.Context, credentials *models.Credentials,
) (*models.Principal, error) {
	log := zerolog.Ctx(ctx)
	if credentials == nil {
		return nil, ErrInvalidCredentials
	}

	var user models.UserDB
	sqlH := database.NewSQLHelper(context.Background(), auth.db, *log)
	if err := sqlH.Get(&user,
		squirrel.
			Select(models.UserDBColumns...).
			From(models.UserDBTable).
			Where(
				squirrel.Eq{models.UserinfoUsernameColumn: credentials.Username}),
	); err != nil {
		if err == sql.ErrNoRows {
			return nil, ErrInvalidCredentials
		}
		return nil, err
	}

	if err := bcrypt.CompareHashAndPassword([]byte(user.Password), []byte(credentials.Password)); err != nil {
		return nil, ErrInvalidCredentials
	}

	return &models.Principal{
		StandardClaims: jwt.StandardClaims{
			Subject: user.Username,
		},
		IsAdmin: user.IsAdmin,
	}, nil
}

// AuthFromAuthorization loads a 'Authorization' header and returns a Principal instance
func (auth *Auth) AuthFromAuthorization(header string) (*models.Principal, error) {
	if !strings.HasPrefix(header, "Bearer ") {
		return nil, openapi_errors.New(
			http.StatusUnauthorized,
			"Invalid Authorization header value. Expects 'Bearer <token>', got '%s'",
			header)
	}
	tokenString := header[7:]
	return auth.AuthFromToken(tokenString)
}

// AuthFromCookie loads a 'Cookie' header and returns a Principal instance
func (auth *Auth) AuthFromCookie(header string) (*models.Principal, error) {
	h := http.Header{}
	h.Add("Cookie", header)
	request := http.Request{Header: h}

	cookie, err := request.Cookie("rednerd-auth")
	if err != nil {
		if err == http.ErrNoCookie {
			return nil, fmt.Errorf("no rednerd-auth cookie found")
		}
		return nil, err
	}
	return auth.AuthFromToken(cookie.Value)
}

// AuthFromAPIKey loads a 'Rednerd-API-Key' header and returns a Principal instance
func (auth *Auth) AuthFromAPIKey(header string) (*models.Principal, error) {
	cacheKey := "apikey:" + header
	if value, ok := auth.cache.Get(cacheKey); ok {
		auth.cacheHitCollector.Inc()
		return value.(*models.Principal), nil
	}
	auth.cacheMissCollector.Inc()
	var userColumns = make([]string, len(models.UserDBColumns))
	for i, c := range models.UserDBColumns {
		userColumns[i] = models.UserDBTable + "." + c
	}
	var user models.UserDB
	sqlH := database.NewSQLHelper(context.Background(), auth.db, auth.log)
	if err := sqlH.Get(&user,
		squirrel.
			Select(userColumns...).
			From(models.UserDBTable).
			LeftJoin(database.Join(
				models.UserDBTable,
				models.UserDBUsernameColumn,
				models.ApikeyDBTable,
				models.ApikeyDBUsernameColumn)).
			Where(
				squirrel.Eq{models.ApikeyDBValueColumn: header}),
	); err != nil {
		if err == sql.ErrNoRows {
			return nil, ErrInvalidCredentials
		}
		return nil, err
	}
	principal := &models.Principal{
		StandardClaims: jwt.StandardClaims{
			Subject: user.Username,
		},
		IsAdmin: user.IsAdmin,
	}
	auth.addToCache(cacheKey, principal)
	return principal, nil
}

// AuthFromToken loads a token
func (auth *Auth) AuthFromToken(tokenString string) (*models.Principal, error) {
	if value, ok := auth.cache.Get(tokenString); ok {
		auth.cacheHitCollector.Inc()
		return value.(*models.Principal), nil
	}
	auth.cacheMissCollector.Inc()

	token, err := auth.parser.ParseWithClaims(
		tokenString,
		&models.Principal{},
		func(token *jwt.Token) (interface{}, error) {
			return auth.secret, nil
		})
	if err != nil {
		return nil, openapi_errors.New(http.StatusUnauthorized, "Invalid token: %s", err)
	}
	if !token.Valid {
		return nil, openapi_errors.New(http.StatusUnauthorized, "Invalid token")
	}
	claims := token.Claims.(*models.Principal)
	auth.addToCache(tokenString, claims)
	return claims, nil
}

// MakeToken creates a new token
func (auth *Auth) MakeToken(principal models.Principal, withCookie bool) (string, *http.Cookie, error) {
	expiresIn := auth.expiration
	if withCookie {
		expiresIn *= 10
	}
	expiresAt := time.Now().Add(expiresIn)
	principal.StandardClaims.ExpiresAt = expiresAt.Unix()
	token := jwt.NewWithClaims(jwt.SigningMethodHS256, principal)
	stoken, err := token.SignedString(auth.secret)
	if err != nil {
		return "", nil, err
	}
	auth.addToCache(stoken, &principal)

	var cookie *http.Cookie
	if withCookie {
		cookie = &http.Cookie{
			Name:    "rednerd-auth",
			Value:   stoken,
			Expires: expiresAt,
			Path:    "/",
			Domain:  auth.cookieDomain,
			Secure:  auth.cookieSecure,
		}
	}
	return stoken, cookie, nil
}
