package cmd

import (
	"bytes"
	"context"
	"fmt"
	"os"

	"github.com/jmoiron/sqlx"
	"golang.org/x/crypto/bcrypt"
	"golang.org/x/crypto/ssh/terminal"
	"orus.io/orus-io/rednerd/database"
	"orus.io/orus-io/rednerd/models"
)

// AdminPasswordCmd is the 'admin-user' command
type AdminPasswordCmd struct {
	options *Options `no-flag:"t"`
}

// Execute reset the admin user password
func (cmd *AdminPasswordCmd) Execute([]string) error {
	db, err := sqlx.Open("postgres", cmd.options.Dsn)
	if err != nil {
		return err
	}
	defer func() {
		if err := db.Close(); err != nil {
			log.Err(err).Msg("could not close database connection properly")
		}
	}()
	if err := db.Ping(); err != nil {
		return err
	}
	tx, err := database.Begin(context.Background(), db)
	if err != nil {
		return err
	}
	defer tx.RollbackIfOpened(log)

	fmt.Print("Type the new admin password: ")
	password, err := terminal.ReadPassword(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}
	fmt.Println()

	fmt.Print("Please confirm: ")
	passwordConfirm, err := terminal.ReadPassword(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}
	fmt.Println()

	if !bytes.Equal(password, passwordConfirm) {
		return fmt.Errorf("passwords mismatched")
	}

	if err := SetAdminPassword(tx.Tx, password); err != nil {
		return err
	}

	return tx.Commit()
}

// SetAdminPassword resets the admin account and its password
func SetAdminPassword(tx *sqlx.Tx, password []byte) error {
	// Salt and hash the password using the bcrypt algorithm
	// The second argument is the cost of hashing, which we arbitrarily set as 8
	// (this value can be more or less, depending on the computing power you wish to utilize)
	hashedPassword, err := bcrypt.GenerateFromPassword(password, 10)
	if err != nil {
		return err
	}

	var (
		account = models.AccountDB{
			Name:   "admin",
			IsUser: true,
		}
		user = models.UserDB{
			Userinfo: models.Userinfo{
				Username: "admin",
				IsAdmin:  true,
			},
			Password: string(hashedPassword),
		}
	)
	h := database.NewSQLHelper(context.Background(), tx, log)

	if err := h.Upsert(account); err != nil {
		return err
	}
	if err := h.Upsert(user); err != nil {
		return err
	}
	return nil
}
