package wpd

import (
	"bufio"
	"context"
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"sync"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"orus.io/orus-io/rednerd/testutils"
)

type MockRunner struct {
	operations *[]Op
	t          *testing.T
	closing    bool

	sendInfo bool

	stdinIn   *io.PipeWriter
	stdinOut  *io.PipeReader
	stdoutIn  *io.PipeWriter
	stdoutOut *io.PipeReader

	wg sync.WaitGroup
}

type ExpectedMsg struct {
	Head string
	Body []byte
}

type Op struct {
	Expects     *ExpectedMsg
	Send        string
	InError     error
	OutError    error
	CreateError error
}

func (mr *MockRunner) Stdin() io.WriteCloser {
	return mr.stdinIn
}

func (mr *MockRunner) Stdout() io.ReadCloser {
	return mr.stdoutOut
}

func (mr *MockRunner) Stop() error {
	mr.closing = true
	mr.stdinIn.Close()
	mr.stdinOut.Close()
	mr.stdoutIn.Close()
	mr.stdoutOut.Close()
	mr.wg.Wait()
	return nil
}

func (mr *MockRunner) run() {
	defer mr.wg.Done()

	if mr.sendInfo {
		fmt.Println("Sending info")
		if _, err := mr.stdoutIn.Write([]byte(
			"{\"wpd\":\"mock\",\"version\":\"1\",\"protocol\":\"1\"}\n",
		)); err != nil {
			mr.t.Error(err)
			return
		}
	}

	fmt.Println("main loop")
	stdinOut := bufio.NewReader(mr.stdinOut)
	for {
		if len(*mr.operations) == 0 {
			for {
				_, err := stdinOut.ReadByte()
				if err != nil {
					return
				}
			}
		}
		op := (*mr.operations)[0]
		*mr.operations = (*mr.operations)[1:]

		if op.Expects != nil {
			line, err := stdinOut.ReadBytes('\n')
			if mr.closing {
				return
			}
			if !assert.NoError(mr.t, err) {
				*mr.operations = nil
				continue
			}
			line = line[:len(line)-1]
			var js map[string]interface{}
			if !assert.NoError(mr.t, json.Unmarshal(line, &js)) ||
				!assert.JSONEq(mr.t, op.Expects.Head, string(line)) {
				fmt.Println("Some assertion failed, stop the operations")
				*mr.operations = nil
				continue
			}
			if op.Expects.Body == nil {
			} else {
				var data = make([]byte, len(op.Expects.Body))
				_, err := io.ReadFull(stdinOut, data)
				if mr.closing {
					*mr.operations = nil
					continue
				}
				assert.NoError(mr.t, err)
				if err != nil {
					*mr.operations = nil
					continue
				}
				if !assert.Equal(mr.t, string(op.Expects.Body), string(data)) {
					*mr.operations = nil
					continue
				}
				b, err := stdinOut.ReadByte()
				if mr.closing {
					*mr.operations = nil
					continue
				}
				if !assert.NoError(mr.t, err) || !assert.Equal(mr.t, byte('\n'), b) {
					*mr.operations = nil
					continue
				}
			}
		} else if op.Send != "" {
			_, err := mr.stdoutIn.Write([]byte(op.Send))
			if mr.closing {
				return
			}
			if !assert.NoError(mr.t, err) {
				*mr.operations = nil
				continue
			}
		}
	}
}

func MockRunnerFactory(t *testing.T, sendInfo bool, operations ...Op) func() (Runner, error) {
	return func() (Runner, error) {
		if len(operations) != 0 && operations[0].CreateError != nil {
			op := operations[0]
			operations = operations[1:]
			return nil, op.CreateError
		}
		inOut, inIn := io.Pipe()
		outOut, outIn := io.Pipe()
		m := MockRunner{
			operations: &operations,
			t:          t,
			sendInfo:   sendInfo,
			stdinIn:    inIn,
			stdinOut:   inOut,
			stdoutIn:   outIn,
			stdoutOut:  outOut,
		}
		m.wg.Add(1)
		go m.run()
		return &m, nil
	}
}

func TestWpd(t *testing.T) {
	t.Run("create error", func(t *testing.T) {
		log := testutils.GetLogger(t)
		_, err := NewWPD(MockRunnerFactory(t, true, Op{CreateError: fmt.Errorf("cannot create")}), log)
		assert.EqualError(t, err, "cannot create")
	})
	t.Run("single job", func(t *testing.T) {
		log := testutils.GetLogger(t)
		ctx, cancel := context.WithTimeout(context.Background(), time.Second*2)
		defer cancel()
		wpd, err := NewWPD(MockRunnerFactory(t, true, Op{Expects: &ExpectedMsg{
			Head: `{"id": 1, "css": "body{}", "size": 16}`,
			Body: []byte("<p>paragraph</p>"),
		}}, Op{Send: `{"id":1,"size":2}
ok
`}), log)
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		defer func() {
			assert.NoError(t, wpd.Close())
		}()

		pdf, err := wpd.Render(ctx, "test", "body{}", []byte("<p>paragraph</p>"))
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		assert.Equal(t, "ok", string(pdf))
	})
	t.Run("Two jobs", func(t *testing.T) {
		log := testutils.GetLogger(t)
		ctx, cancel := context.WithTimeout(context.Background(), time.Second*2)
		defer cancel()
		wpd, err := NewWPD(MockRunnerFactory(t, true,
			Op{Expects: &ExpectedMsg{
				Head: `{"id": 1, "css": "body{}", "size": 16}`,
				Body: []byte("<p>paragraph</p>"),
			}},
			Op{Send: `{"id":1,"size":2}
ok
`},
			Op{Expects: &ExpectedMsg{
				Head: `{"id": 2, "css": "body{}", "size": 9}`,
				Body: []byte("<p>p2</p>"),
			}},
			Op{Send: `{"id":2,"size":3}
ok2
`},
		), log)
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		defer func() {
			assert.NoError(t, wpd.Close())
		}()

		pdf, err := wpd.Render(ctx, "test", "body{}", []byte("<p>paragraph</p>"))
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		assert.Equal(t, "ok", string(pdf))

		pdf, err = wpd.Render(ctx, "test", "body{}", []byte("<p>p2</p>"))
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		assert.Equal(t, "ok2", string(pdf))
	})
	t.Run("interleaved jobs", func(t *testing.T) {
		log := testutils.GetLogger(t)
		ctx, cancel := context.WithTimeout(context.Background(), time.Second*2)
		defer cancel()
		wpd, err := NewWPD(MockRunnerFactory(t, true,
			Op{Expects: &ExpectedMsg{
				Head: `{"id": 1, "css": "body{}", "size": 16}`,
				Body: []byte("<p>paragraph</p>"),
			}},
			Op{Expects: &ExpectedMsg{
				Head: `{"id": 2, "css": "body{}", "size": 9}`,
				Body: []byte("<p>p2</p>"),
			}},
			Op{Send: `{"id":2,"size":3}
ok2
`},
			Op{Send: `{"id":1,"size":2}
ok
`},
		), log)
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		defer func() {
			assert.NoError(t, wpd.Close())
		}()

		job1 := wpd.Submit(ctx, "test", "body{}", []byte("<p>paragraph</p>"))

		job2 := wpd.Submit(ctx, "test", "body{}", []byte("<p>p2</p>"))

		res1 := <-job1.Result
		if !assert.NoError(t, res1.Error) {
			t.FailNow()
		}
		assert.Equal(t, "ok", string(res1.Content))

		res2 := <-job2.Result
		if !assert.NoError(t, res1.Error) {
			t.FailNow()
		}
		assert.Equal(t, "ok2", string(res2.Content))
	})

	t.Run("cancel job", func(t *testing.T) {
		log := testutils.GetLogger(t)
		ctx, cancel := context.WithTimeout(context.Background(), time.Second*2)
		defer cancel()
		wpd, err := NewWPD(MockRunnerFactory(t, true,
			Op{Expects: &ExpectedMsg{
				Head: `{"id": 1, "css": "body{}", "size": 16}`,
				Body: []byte("<p>paragraph</p>"),
			}},
			Op{Expects: &ExpectedMsg{
				Head: `{"id": 1, "action": "cancel"}`,
			}},
		), log)
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		defer func() {
			assert.NoError(t, wpd.Close())
		}()

		job1 := wpd.Submit(ctx, "test", "body{}", []byte("<p>paragraph</p>"))
		cancel()

		res1 := <-job1.Result
		assert.EqualError(t, res1.Error, "context canceled")
	})
}

func TestRealWpd(t *testing.T) {
	log := testutils.GetLogger(t)
	ctx, cancel := context.WithTimeout(context.Background(), time.Second*5)
	defer cancel()
	wpd, err := NewWPD(RealRunnerFactory, log)
	if !assert.NoError(t, err) {
		t.FailNow()
	}
	c, err := wpd.Render(ctx, "test", "", []byte("<p>A small paragraph</p>"))
	if !assert.NoError(t, err) {
		t.FailNow()
	}
	assert.NoError(t, ioutil.WriteFile(t.Name()+".pdf", c, 0600))
}
