package handlers

import (
	"context"
	"database/sql"
	"errors"
	"strings"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/rednerd/database"
	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/template_management"
	"orus.io/orus-io/rednerd/utils"
)

// NewAccountTemplateUpdateHandler returns a AccountTemplateUpdateHandler
func NewAccountTemplateUpdateHandler(db *sqlx.DB) *AccountTemplateUpdateHandler {
	var columns = make([]string, len(models.TemplateDBDataColumns)-1)
	// We don't want the 'internal_version' column when updating.
	copy(columns[:11], models.TemplateDBDataColumns[:11])
	copy(columns[11:], models.TemplateDBDataColumns[12:])
	return &AccountTemplateUpdateHandler{
		db:      db,
		columns: columns,
	}
}

// AccountTemplateUpdateHandler handles GET /user
type AccountTemplateUpdateHandler struct {
	db      *sqlx.DB
	columns []string
}

func (h *AccountTemplateUpdateHandler) handle(
	ctx context.Context, account, name string, template models.Template,
) (*models.Template, error) {
	var (
		log = zerolog.Ctx(ctx)
	)

	templatedb := models.TemplateDB{Template: template}
	if err := templatedb.ToDB(); err != nil {
		return nil, err
	}

	q := squirrel.Update(models.TemplateDBTable).
		SetMap(database.ValuesMap(templatedb, h.columns...)).
		Where(squirrel.Eq{
			models.TemplateDBAccountColumn: account,
			models.TemplateDBNameColumn:    name,
		}).
		Suffix("RETURNING " + strings.Join(models.TemplateDBColumns, ","))

	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return nil, err
	}
	defer tx.RollbackIfOpened(*log)
	if err := database.GetContext(ctx, tx, &templatedb, q, log); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return nil, utils.NewHTTPErrorf(404, "No such template")
		}
		return nil, err
	}
	templatedb.FromDB()
	templatedb.ForceBase64()
	if err := tx.Commit(); err != nil {
		return nil, err
	}
	return &templatedb.Template, nil
}

// Handle handles the incoming request
func (h *AccountTemplateUpdateHandler) Handle(
	params template_management.AccountTemplateUpdateParams, principal *models.Principal,
) middleware.Responder {
	if principal.IsAdmin {
		return template_management.
			NewAccountTemplateAddUnauthorized().
			WithPayload(&models.Error{
				Message: "Admin cannot manage templates. Please use a non-admin user",
			})
	}
	if principal.Subject != params.Account {
		return template_management.
			NewAccountTemplateAddUnauthorized().
			WithPayload(&models.Error{
				Message: "User can manage its own templates only",
			})
	}
	if params.Template.Account == "" {
		params.Template.Account = params.Account
	}
	if params.Template.Name == "" {
		params.Template.Name = params.Name
	}
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)
	template, err := h.handle(
		params.HTTPRequest.Context(),
		params.Account,
		params.Name,
		*params.Template)
	if err != nil {
		if httpErr, ok := err.(utils.HTTPError); ok {
			return HTTPErrorResponder{httpErr}
		}
		return InternalError(err, log, false)
	}
	return template_management.NewAccountTemplateUpdateOK().WithPayload(template)
}
