package handlers

import (
	"context"
	"errors"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/rednerd/database"
	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/user_management"
)

// ErrNoSuchUserApikey is returned if the user does not exist
var ErrNoSuchUserApikey = errors.New("no such user")

// NewUserApikeyDeleteHandler returns a UserApikeyDeleteHandler
func NewUserApikeyDeleteHandler(db *sqlx.DB) *UserApikeyDeleteHandler {
	return &UserApikeyDeleteHandler{
		db: db,
	}
}

// UserApikeyDeleteHandler handles GET /user
type UserApikeyDeleteHandler struct {
	db *sqlx.DB
}

func (h *UserApikeyDeleteHandler) handle(
	ctx context.Context, username string, name string,
) error {
	var (
		log = zerolog.Ctx(ctx)
	)
	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return err
	}
	defer tx.RollbackIfOpened(*log)
	r, err := database.ExecContext(ctx, tx,
		squirrel.Delete(models.ApikeyDBTable).
			From(models.ApikeyDBTable).
			Where(squirrel.Eq{
				models.ApikeyDBUsernameColumn: username,
				models.ApikeyNameColumn:       name,
			}),
		log,
	)
	if err != nil {
		return err
	}
	count, err := r.RowsAffected()
	if err != nil {
		return err
	}
	if count == 0 {
		return ErrNoSuchUserApikey
	}
	return tx.Commit()
}

// Handle handles the incoming request
func (h *UserApikeyDeleteHandler) Handle(
	params user_management.UserApikeyDeleteParams, principal *models.Principal,
) middleware.Responder {
	if !principal.IsAdmin {
		return user_management.NewUserApikeyDeleteUnauthorized()
	}
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)
	err := h.handle(params.HTTPRequest.Context(), params.Username, params.Name)
	switch err {
	case nil:
		return user_management.NewUserApikeyDeleteNoContent()
	case ErrNoSuchUserApikey:
		return user_management.NewUserApikeyDeleteNotFound()
	default:
		return InternalError(err, log, false)
	}
}
