package handlers

import (
	"context"
	"net/http"
	"strings"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/rednerd/database"
	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/user_management"
)

// NewUserUpdateHandler returns a UserUpdateHandler
func NewUserUpdateHandler(db *sqlx.DB) *UserUpdateHandler {
	return &UserUpdateHandler{
		db: db,
	}
}

// UserUpdateHandler handles GET /user
type UserUpdateHandler struct {
	db *sqlx.DB
}

func (h *UserUpdateHandler) handle(
	ctx context.Context, username string, user models.Userinfo,
) (*models.Userinfo, error) {
	var (
		log = zerolog.Ctx(ctx)
	)
	if user.Username == "" {
		user.Username = username
	}

	q := squirrel.Update(models.UserDBTable).
		Set(models.UserinfoUsernameColumn, user.Username).
		Set(models.UserinfoIsAdminColumn, user.IsAdmin).
		Where(squirrel.Eq{models.UserinfoUsernameColumn: username}).
		Suffix("RETURNING " + strings.Join(models.UserinfoColumns, ","))

	var info models.Userinfo
	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return nil, err
	}
	defer tx.RollbackIfOpened(*log)
	if err := database.GetContext(ctx, tx, &info, q, log); err != nil {
		return nil, err
	}
	if err := tx.Commit(); err != nil {
		return nil, err
	}
	return &info, nil
}

// Handle handles the incoming request
func (h *UserUpdateHandler) Handle(
	params user_management.UserUpdateParams, principal *models.Principal,
) middleware.Responder {
	if !principal.IsAdmin {
		return user_management.NewUserUpdateUnauthorized()
	}
	ctx := params.HTTPRequest.Context()
	log := zerolog.Ctx(ctx)
	if params.User == nil {
		return user_management.NewUserUpdateDefault(http.StatusBadRequest)
	}
	user, err := h.handle(params.HTTPRequest.Context(), params.Username, *params.User)
	if err != nil {
		return InternalError(err, log, false)
	}
	return user_management.NewUserUpdateOK().WithPayload(user)
}
