package tests

import (
	"encoding/base64"
	"net/http"
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"orus.io/orus-io/rednerd/models"
)

func TestRender(t *testing.T) {
	tester := NewAPITester(t)
	defer tester.Close()

	defer tester.Logout()
	tester.Login("admin", "admin")

	tester.CreateUser("janedoe", "janedoe")
	tester.Login("janedoe", "janedoe")

	tester.Run("MJML-Template-to-html", func(t *testing.T) {
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "text/html",
				"template": JSONObj{
					"language":    "mustache",
					"produces":    "text/mjml",
					"body-format": "text",
					"body":        "<mjml> <mj-body> <mj-section> <mj-column> <mj-text>hello {{name}}</mj-text> </mj-column> </mj-section> </mj-body> </mjml>"},
				"data": []JSONObj{{"name": "Nath"}},
			},
			&result)
		assert.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello Nath")
		}
	})

	tester.Run("MJML-Saved-Template-to-html", func(t *testing.T) {
		require.Equal(t, http.StatusCreated,
			tester.MustPostJSON(
				"/api/v1/template/janedoe",
				JSONObj{
					"name":        "t1",
					"language":    "mustache",
					"produces":    "text/mjml",
					"body-format": "text",
					"body":        "<mjml> <mj-body> <mj-section> <mj-column> <mj-text>hello {{name}}</mj-text> </mj-column> </mj-section> </mj-body> </mjml>"},
				nil),
		)
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "text/html",
				"template": JSONObj{
					"account": "janedoe",
					"name":    "t1",
				},
				"data": []JSONObj{{"name": "Nath"}},
			},
			&result)
		assert.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello Nath")
		}
	})

	tester.Run("MJML-with-errors", func(t *testing.T) {
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "text/html",
				"document": JSONObj{
					"type":        "text/mjml",
					"body-format": "test",
					"body": `<mjml>
							  <mj-body>
								<mj-section invalid="hi">
								  <mj-column>
									<mj-text invalid=true>hello world</mj-text>
								  </mj-column>
								</mj-section>
							  </mj-body>
							</mjml>`,
				},
			},
			&result,
		)
		require.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello world")
		}
		require.Equal(t, 2, len(result[0].RenderErrors))
		assert.Equal(t, "MJML", result[0].RenderErrors[0].Engine)
		assert.Equal(t, "MJML", result[0].RenderErrors[1].Engine)
	})
	tester.Run("MJML-skip-validation", func(t *testing.T) {
		var result []models.Document
		tester.MustPostJSON(
			"/api/v1/render",
			JSONObj{
				"accept": "text/html",
				"metadata": JSONObj{
					"mjml-validation-level": "skip",
				},
				"document": JSONObj{
					"type":        "text/mjml",
					"body-format": "test",
					"body": `<mjml>
							  <mj-body>
								<mj-section invalid="hi">
								  <mj-column>
									<mj-text invalid=true>hello world</mj-text>
								  </mj-column>
								</mj-section>
							  </mj-body>
							</mjml>`,
				},
			},
			&result,
		)
		require.Equal(t, 1, len(result))
		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello world")
		}
		require.Equal(t, 0, len(result[0].RenderErrors))
	})
}
